@@ -13,7 +13,7 @@ from certbot import account
 from certbot import errors
 from certbot import util
 
-from certbot.tests import test_util
+import certbot.tests.util as test_util
 
 
 KEY = test_util.load_vector("rsa512_key.pem")
@@ -44,31 +44,49 @@ class RegisterTest(unittest.TestCase):
     def test_no_tos(self):
         with mock.patch("certbot.client.acme_client.Client") as mock_client:
             mock_client.register().terms_of_service = "http://tos"
-            with mock.patch("certbot.account.report_new_account"):
-                self.tos_cb.return_value = False
-                self.assertRaises(errors.Error, self._call)
+            with mock.patch("certbot.eff.handle_subscription") as mock_handle:
+                with mock.patch("certbot.account.report_new_account"):
+                    self.tos_cb.return_value = False
+                    self.assertRaises(errors.Error, self._call)
+                    self.assertFalse(mock_handle.called)
 
-                self.tos_cb.return_value = True
-                self._call()
+                    self.tos_cb.return_value = True
+                    self._call()
+                    self.assertTrue(mock_handle.called)
 
-                self.tos_cb = None
-                self._call()
+                    self.tos_cb = None
+                    self._call()
+                    self.assertEqual(mock_handle.call_count, 2)
 
     def test_it(self):
         with mock.patch("certbot.client.acme_client.Client"):
             with mock.patch("certbot.account.report_new_account"):
-                self._call()
+                with mock.patch("certbot.eff.handle_subscription"):
+                    self._call()
 
     @mock.patch("certbot.account.report_new_account")
     @mock.patch("certbot.client.display_ops.get_email")
     def test_email_retry(self, _rep, mock_get_email):
         from acme import messages
+        self.config.noninteractive_mode = False
         msg = "DNS problem: NXDOMAIN looking up MX for example.com"
-        mx_err = messages.Error(detail=msg, typ="urn:acme:error:invalidEmail")
+        mx_err = messages.Error.with_code('invalidContact', detail=msg)
         with mock.patch("certbot.client.acme_client.Client") as mock_client:
-            mock_client().register.side_effect = [mx_err, mock.MagicMock()]
-            self._call()
-            self.assertEqual(mock_get_email.call_count, 1)
+            with mock.patch("certbot.eff.handle_subscription") as mock_handle:
+                mock_client().register.side_effect = [mx_err, mock.MagicMock()]
+                self._call()
+                self.assertEqual(mock_get_email.call_count, 1)
+                self.assertTrue(mock_handle.called)
+
+    @mock.patch("certbot.account.report_new_account")
+    def test_email_invalid_noninteractive(self, _rep):
+        from acme import messages
+        msg = "DNS problem: NXDOMAIN looking up MX for example.com"
+        mx_err = messages.Error.with_code('invalidContact', detail=msg)
+        with mock.patch("certbot.client.acme_client.Client") as mock_client:
+            with mock.patch("certbot.eff.handle_subscription"):
+                mock_client().register.side_effect = [mx_err, mock.MagicMock()]
+                self.assertRaises(errors.Error, self._call)
 
     def test_needs_email(self):
         self.config.email = None
@@ -76,31 +94,38 @@ class RegisterTest(unittest.TestCase):
 
     @mock.patch("certbot.client.logger")
     def test_without_email(self, mock_logger):
-        with mock.patch("certbot.client.acme_client.Client"):
-            with mock.patch("certbot.account.report_new_account"):
-                self.config.email = None
-                self.config.register_unsafely_without_email = True
-                self.config.dry_run = False
-                self._call()
-                mock_logger.warn.assert_called_once_with(mock.ANY)
+        with mock.patch("certbot.eff.handle_subscription") as mock_handle:
+            with mock.patch("certbot.client.acme_client.Client"):
+                with mock.patch("certbot.account.report_new_account"):
+                    self.config.email = None
+                    self.config.register_unsafely_without_email = True
+                    self.config.dry_run = False
+                    self._call()
+                    mock_logger.warning.assert_called_once_with(mock.ANY)
+                    self.assertTrue(mock_handle.called)
 
     def test_unsupported_error(self):
         from acme import messages
         msg = "Test"
         mx_err = messages.Error(detail=msg, typ="malformed", title="title")
         with mock.patch("certbot.client.acme_client.Client") as mock_client:
-            mock_client().register.side_effect = [mx_err, mock.MagicMock()]
-            self.assertRaises(messages.Error, self._call)
+            with mock.patch("certbot.eff.handle_subscription") as mock_handle:
+                mock_client().register.side_effect = [mx_err, mock.MagicMock()]
+                self.assertRaises(messages.Error, self._call)
+        self.assertFalse(mock_handle.called)
 
-class ClientTest(unittest.TestCase):
-    """Tests for certbot.client.Client."""
 
+class ClientTestCommon(unittest.TestCase):
+    """Common base class for certbot.client.Client tests."""
     def setUp(self):
         self.config = mock.MagicMock(
-            no_verify_ssl=False, config_dir="/etc/letsencrypt", allow_subset_of_names=False)
+            no_verify_ssl=False,
+            config_dir="/etc/letsencrypt",
+            work_dir="/var/lib/letsencrypt",
+            allow_subset_of_names=False)
+
         # pylint: disable=star-args
         self.account = mock.MagicMock(**{"key.pem": KEY})
-        self.eg_domains = ["example.com", "www.example.com"]
 
         from certbot.client import Client
         with mock.patch("certbot.client.acme_client.Client") as acme:
@@ -110,6 +135,16 @@ class ClientTest(unittest.TestCase):
                 config=self.config, account_=self.account,
                 auth=None, installer=None)
 
+
+class ClientTest(ClientTestCommon):
+    """Tests for certbot.client.Client."""
+    def setUp(self):
+        super(ClientTest, self).setUp()
+
+        self.config.allow_subset_of_names = False
+        self.config.config_dir = "/etc/letsencrypt"
+        self.eg_domains = ["example.com", "www.example.com"]
+
     def test_init_acme_verify_ssl(self):
         net = self.acme_client.call_args[1]["net"]
         self.assertTrue(net.verify_ssl)
@@ -203,6 +238,27 @@ class ClientTest(unittest.TestCase):
             mock.sentinel.key, domains, self.config.csr_dir)
         self._check_obtain_certificate()
 
+    @mock.patch('certbot.client.Client.obtain_certificate')
+    @mock.patch('certbot.storage.RenewableCert.new_lineage')
+    @mock.patch('OpenSSL.crypto.dump_certificate')
+    def test_obtain_and_enroll_certificate(self, mock_dump_certificate,
+        mock_storage, mock_obtain_certificate):
+        domains = ["example.com", "www.example.com"]
+        mock_obtain_certificate.return_value = (mock.MagicMock(),
+            mock.MagicMock(), mock.MagicMock(), None)
+
+        self.client.config.dry_run = False
+        self.assertTrue(self.client.obtain_and_enroll_certificate(domains, "example_cert"))
+
+        self.assertTrue(self.client.obtain_and_enroll_certificate(domains, None))
+
+        self.client.config.dry_run = True
+
+        self.assertFalse(self.client.obtain_and_enroll_certificate(domains, None))
+
+        self.assertTrue(mock_storage.call_count == 2)
+        self.assertTrue(mock_dump_certificate.call_count == 2)
+
     @mock.patch("certbot.cli.helpful_parser")
     def test_save_certificate(self, mock_parser):
         # pylint: disable=too-many-locals
@@ -232,11 +288,11 @@ class ClientTest(unittest.TestCase):
         self.assertEqual(os.path.dirname(fullchain_path),
                          os.path.dirname(candidate_fullchain_path))
 
-        with open(cert_path, "r") as cert_file:
+        with open(cert_path, "rb") as cert_file:
             cert_contents = cert_file.read()
         self.assertEqual(cert_contents, test_util.load_vector(certs[0]))
 
-        with open(chain_path, "r") as chain_file:
+        with open(chain_path, "rb") as chain_file:
             chain_contents = chain_file.read()
         self.assertEqual(chain_contents, test_util.load_vector(certs[1]) +
                          test_util.load_vector(certs[2]))
@@ -279,7 +335,7 @@ class ClientTest(unittest.TestCase):
                           ["foo.bar"], "key", "cert", "chain", "fullchain")
         installer.recovery_routine.assert_called_once_with()
 
-    @mock.patch("certbot.client.zope.component.getUtility")
+    @test_util.patch_get_utility()
     def test_deploy_certificate_restart_failure(self, mock_get_utility):
         installer = mock.MagicMock()
         installer.restart.side_effect = [errors.PluginError, None]
@@ -291,7 +347,7 @@ class ClientTest(unittest.TestCase):
         installer.rollback_checkpoints.assert_called_once_with()
         self.assertEqual(installer.restart.call_count, 2)
 
-    @mock.patch("certbot.client.zope.component.getUtility")
+    @test_util.patch_get_utility()
     def test_deploy_certificate_restart_failure2(self, mock_get_utility):
         installer = mock.MagicMock()
         installer.restart.side_effect = errors.PluginError
@@ -304,137 +360,109 @@ class ClientTest(unittest.TestCase):
         installer.rollback_checkpoints.assert_called_once_with()
         self.assertEqual(installer.restart.call_count, 1)
 
-    @mock.patch("certbot.client.enhancements")
-    def test_enhance_config(self, mock_enhancements):
-        config = ConfigHelper(redirect=True, hsts=False, uir=False)
-        self.assertRaises(errors.Error,
-                          self.client.enhance_config, ["foo.bar"], config)
 
-        mock_enhancements.ask.return_value = True
-        installer = mock.MagicMock()
-        self.client.installer = installer
-        installer.supported_enhancements.return_value = ["redirect"]
+class EnhanceConfigTest(ClientTestCommon):
+    """Tests for certbot.client.Client.enhance_config."""
+    def setUp(self):
+        super(EnhanceConfigTest, self).setUp()
 
-        self.client.enhance_config(["foo.bar"], config)
-        installer.enhance.assert_called_once_with("foo.bar", "redirect", None)
-        self.assertEqual(installer.save.call_count, 1)
-        installer.restart.assert_called_once_with()
+        self.config.hsts = False
+        self.config.redirect = False
+        self.config.staple = False
+        self.config.uir = False
+        self.domain = "example.org"
+
+    def test_no_installer(self):
+        self.assertRaises(
+            errors.Error, self.client.enhance_config, [self.domain], None)
 
     @mock.patch("certbot.client.enhancements")
-    def test_enhance_config_no_ask(self, mock_enhancements):
-        config = ConfigHelper(redirect=True, hsts=False, uir=False)
-        self.assertRaises(errors.Error,
-                          self.client.enhance_config, ["foo.bar"], config)
+    def test_unsupported(self, mock_enhancements):
+        self.client.installer = mock.MagicMock()
+        self.client.installer.supported_enhancements.return_value = []
 
-        mock_enhancements.ask.return_value = True
-        installer = mock.MagicMock()
-        self.client.installer = installer
-        installer.supported_enhancements.return_value = ["redirect", "ensure-http-header"]
-
-        config = ConfigHelper(redirect=True, hsts=False, uir=False)
-        self.client.enhance_config(["foo.bar"], config)
-        installer.enhance.assert_called_with("foo.bar", "redirect", None)
-
-        config = ConfigHelper(redirect=False, hsts=True, uir=False)
-        self.client.enhance_config(["foo.bar"], config)
-        installer.enhance.assert_called_with("foo.bar", "ensure-http-header",
-                "Strict-Transport-Security")
-
-        config = ConfigHelper(redirect=False, hsts=False, uir=True)
-        self.client.enhance_config(["foo.bar"], config)
-        installer.enhance.assert_called_with("foo.bar", "ensure-http-header",
-                "Upgrade-Insecure-Requests")
-
-        self.assertEqual(installer.save.call_count, 3)
-        self.assertEqual(installer.restart.call_count, 3)
-
-    @mock.patch("certbot.client.enhancements")
-    def test_enhance_config_unsupported(self, mock_enhancements):
-        installer = mock.MagicMock()
-        self.client.installer = installer
-        installer.supported_enhancements.return_value = []
-
-        config = ConfigHelper(redirect=None, hsts=True, uir=True)
-        self.client.enhance_config(["foo.bar"], config)
-        installer.enhance.assert_not_called()
+        self.config.redirect = None
+        self.config.hsts = True
+        with mock.patch("certbot.client.logger") as mock_logger:
+            self.client.enhance_config([self.domain], None)
+        self.assertEqual(mock_logger.warning.call_count, 1)
+        self.client.installer.enhance.assert_not_called()
         mock_enhancements.ask.assert_not_called()
 
-    def test_enhance_config_no_installer(self):
-        config = ConfigHelper(redirect=True, hsts=False, uir=False)
-        self.assertRaises(errors.Error,
-                          self.client.enhance_config, ["foo.bar"], config)
+    def test_no_ask_hsts(self):
+        self.config.hsts = True
+        self._test_with_all_supported()
+        self.client.installer.enhance.assert_called_with(
+            self.domain, "ensure-http-header", "Strict-Transport-Security")
 
-    @mock.patch("certbot.client.zope.component.getUtility")
-    @mock.patch("certbot.client.enhancements")
-    def test_enhance_config_enhance_failure(self, mock_enhancements,
-                                            mock_get_utility):
-        mock_enhancements.ask.return_value = True
+    def test_no_ask_redirect(self):
+        self.config.redirect = True
+        self._test_with_all_supported()
+        self.client.installer.enhance.assert_called_with(
+            self.domain, "redirect", None)
+
+    def test_no_ask_staple(self):
+        self.config.staple = True
+        self._test_with_all_supported()
+        self.client.installer.enhance.assert_called_with(
+            self.domain, "staple-ocsp", None)
+
+    def test_no_ask_uir(self):
+        self.config.uir = True
+        self._test_with_all_supported()
+        self.client.installer.enhance.assert_called_with(
+            self.domain, "ensure-http-header", "Upgrade-Insecure-Requests")
+
+    def test_enhance_failure(self):
+        self.client.installer = mock.MagicMock()
+        self.client.installer.enhance.side_effect = errors.PluginError
+        self._test_error()
+        self.client.installer.recovery_routine.assert_called_once_with()
+
+    def test_save_failure(self):
+        self.client.installer = mock.MagicMock()
+        self.client.installer.save.side_effect = errors.PluginError
+        self._test_error()
+        self.client.installer.recovery_routine.assert_called_once_with()
+        self.client.installer.save.assert_called_once_with(mock.ANY)
+
+    def test_restart_failure(self):
+        self.client.installer = mock.MagicMock()
+        self.client.installer.restart.side_effect = [errors.PluginError, None]
+        self._test_error_with_rollback()
+
+    def test_restart_failure2(self):
         installer = mock.MagicMock()
-        self.client.installer = installer
-        installer.supported_enhancements.return_value = ["redirect"]
-        installer.enhance.side_effect = errors.PluginError
-
-        config = ConfigHelper(redirect=True, hsts=False, uir=False)
-
-        self.assertRaises(errors.PluginError,
-                          self.client.enhance_config, ["foo.bar"], config)
-        installer.recovery_routine.assert_called_once_with()
-        self.assertEqual(mock_get_utility().add_message.call_count, 1)
-
-    @mock.patch("certbot.client.zope.component.getUtility")
-    @mock.patch("certbot.client.enhancements")
-    def test_enhance_config_save_failure(self, mock_enhancements,
-                                         mock_get_utility):
-        mock_enhancements.ask.return_value = True
-        installer = mock.MagicMock()
-        self.client.installer = installer
-        installer.supported_enhancements.return_value = ["redirect"]
-        installer.save.side_effect = errors.PluginError
-
-        config = ConfigHelper(redirect=True, hsts=False, uir=False)
-
-        self.assertRaises(errors.PluginError,
-                          self.client.enhance_config, ["foo.bar"], config)
-        installer.recovery_routine.assert_called_once_with()
-        self.assertEqual(mock_get_utility().add_message.call_count, 1)
-
-    @mock.patch("certbot.client.zope.component.getUtility")
-    @mock.patch("certbot.client.enhancements")
-    def test_enhance_config_restart_failure(self, mock_enhancements,
-                                            mock_get_utility):
-        mock_enhancements.ask.return_value = True
-        installer = mock.MagicMock()
-        self.client.installer = installer
-        installer.supported_enhancements.return_value = ["redirect"]
-        installer.restart.side_effect = [errors.PluginError, None]
-
-        config = ConfigHelper(redirect=True, hsts=False, uir=False)
-
-        self.assertRaises(errors.PluginError,
-                          self.client.enhance_config, ["foo.bar"], config)
-
-        self.assertEqual(mock_get_utility().add_message.call_count, 1)
-        installer.rollback_checkpoints.assert_called_once_with()
-        self.assertEqual(installer.restart.call_count, 2)
-
-    @mock.patch("certbot.client.zope.component.getUtility")
-    @mock.patch("certbot.client.enhancements")
-    def test_enhance_config_restart_failure2(self, mock_enhancements,
-                                             mock_get_utility):
-        mock_enhancements.ask.return_value = True
-        installer = mock.MagicMock()
-        self.client.installer = installer
-        installer.supported_enhancements.return_value = ["redirect"]
         installer.restart.side_effect = errors.PluginError
         installer.rollback_checkpoints.side_effect = errors.ReverterError
+        self.client.installer = installer
+        self._test_error_with_rollback()
 
-        config = ConfigHelper(redirect=True, hsts=False, uir=False)
+    @mock.patch("certbot.client.enhancements.ask")
+    def test_ask(self, mock_ask):
+        self.config.redirect = None
+        mock_ask.return_value = True
+        self._test_with_all_supported()
 
-        self.assertRaises(errors.PluginError,
-                          self.client.enhance_config, ["foo.bar"], config)
-        self.assertEqual(mock_get_utility().add_message.call_count, 1)
-        installer.rollback_checkpoints.assert_called_once_with()
-        self.assertEqual(installer.restart.call_count, 1)
+    def _test_error_with_rollback(self):
+        self._test_error()
+        self.assertTrue(self.client.installer.restart.called)
+
+    def _test_error(self):
+        self.config.redirect = True
+        with test_util.patch_get_utility() as mock_gu:
+            self.assertRaises(
+                errors.PluginError, self._test_with_all_supported)
+        self.assertEqual(mock_gu().add_message.call_count, 1)
+
+    def _test_with_all_supported(self):
+        if self.client.installer is None:
+            self.client.installer = mock.MagicMock()
+        self.client.installer.supported_enhancements.return_value = [
+            "ensure-http-header", "redirect", "staple-ocsp"]
+        self.client.enhance_config([self.domain], None)
+        self.assertEqual(self.client.installer.save.call_count, 1)
+        self.assertEqual(self.client.installer.restart.call_count, 1)
 
 
 class RollbackTest(unittest.TestCase):
